% An implementation of direct collocation to open loop 
% dynamic optimisation of a two plant 
% Energy Storage System using CasADi
%% VARIABLE DEMAND!!!

% Zawadi Mdoe, 2018
% ============================================================================
clear;
clc;

addpath('C:\Users\DELL\Desktop\Matlab\casadi-windows-matlabR2016a-v3.4.5')
import casadi.*

run Parameters_EnergyStorage.m
% load CaliforniaDemand.mat
load Expecteddemand.mat

% Degree of interpolating polynomial
d = 3;

% Get collocation points
tau_root = [0 collocation_points(d, 'radau')]; %can be 'legendre'

% Coefficients of the collocation equation
C = zeros(d+1,d+1);

% Coefficients of the continuity equation
D = zeros(d+1, 1);

% Coefficients of the quadrature function
B = zeros(d+1, 1);

% Construct polynomial basis
for j=1:d+1
  % Construct Lagrange polynomials to get the polynomial basis 
  % at the collocation point
  coeff = 1;
  for r=1:d+1
    if r ~= j
      coeff = conv(coeff, [1, -tau_root(r)]);
      coeff = coeff / (tau_root(j)-tau_root(r));
    end
  end
  % Evaluate the polynomial at the final time to get the 
  % coefficients of the continuity equation
  D(j) = polyval(coeff, 1.0);

  % Evaluate the time derivative of the polynomial at all collocation 
  % points to get the coefficients of the continuity equation
  pder = polyder(coeff);
  for r=1:d+1
    C(j,r) = polyval(pder, tau_root(r));
  end

  % Evaluate the integral of the polynomial to get the coefficients 
  % of the quadrature function
  pint = polyint(coeff);
  B(j) = polyval(pint, 1.0);
end

% Time horizon
T = 24*60*60;
ndiff = 1;                 %number of differential states
nalg = 4;                 %number of algebraic states
nu = 3;                 %number of controls
nx = nalg + ndiff;           %total number of states

zub = Inf*ones(nalg,1);
zlb = -Inf*ones(nalg,1);

% Declare model variables
x1 = SX.sym('x1');
x2 = SX.sym('x2');
x3 = SX.sym('x3');
x4 = SX.sym('x6');
x = [x1;x2]
z = [x3; x4];
u1 = SX.sym('u1');
u2 = SX.sym('u2');
u = [u1; u2];

% Model equations

xdot = [(u1*(x3-x1) - (U_hex*A_hex/(rho*cp))*(0.5*(x4+x5))^(1/n))/V_hex;
    
        (q_R2*(T2-x2) + (U_hex*A_hex/(rho*cp))*(0.5*(x4+x5))^(1/n))/V_hex;        
        
        (q_L1*(T1-T_L1)-u1*(x3-x1) + (Q-h_s*A_tank*(x3-T_s))/rho*cp)/V_tank;
        
        x4^(1/n) - x1 + T2;
        
        x5^(1/n) - x3 + x2];

% Objective term
L = Pm*u3 + Pt*u2;

% Continuous time dynamics
f = Function('f', {x, z, u}, {xdot, L});

% Control discretization
N = 24; % number of control intervals
M = 24; % number of MPC loops
h = T/N;
period = N;


%% Prepare output variables
x_opt = zeros(ndiff,M);
u_opt = zeros(nu,M);
i_infeasible = zeros(M,1);
solver_return = {};

%% Predicted Demand
% Q_demand = [CalDemand; CalDemand];

%% MPC loop
for i=1:M
    % Start with an empty NLP
    w={};
    w0 = [];
    lbw = [];
    ubw = [];
    J = 0;
    g={};
    lbg = [];
    ubg = [];

    % "Lift" initial conditions
    % Differential states
    Xk = MX.sym('X0', ndiff);
    w = {w{:}, Xk};
    lbw = [lbw; x_init];
    ubw = [ubw; x_init];
    w0 = [w0; x_init];
    
    % Algebraic states
    Zk = MX.sym('Z0', nalg);
    w = {w{:}, Zk};
    lbw = [lbw; z_init];
    ubw = [ubw; z_init];
    w0 = [w0; z_init];
    
    % Formulate the NLP
    for k=0:N-1
        % New NLP variable for the control
        Uk = MX.sym(['U_' num2str(k)], nu); 
        w = {w{:}, Uk};
        lbw = [lbw; ulb];
        ubw = [ubw; uub];
        w0 = [w0; u_init];
  
        % State at collocation points
        Xkj = {};
        Zkj = {};
        for j=1:d
            % Differential states
            Xkj{j} = MX.sym(['X_' num2str(k) '_' num2str(j)], ndiff);
            w = {w{:}, Xkj{j}};
            lbw = [lbw; xlb];
            ubw = [ubw; xub];
            w0 = [w0; x_init];
            
            % Algebraic states
            Zkj{j} = MX.sym(['Z_' num2str(k) '_' num2str(j)], nalg);
            w = {w{:}, Zkj{j}};
            lbw = [lbw; zlb];
            ubw = [ubw; zub];
            w0 = [w0; z_init];           
        end

        % Loop over collocation points
        Xk_end = D(1)*Xk;
       
        for j=1:d
            % Expression for the state derivative at the collocation point
            xp = C(1,j+1)*Xk;
            zp = tol*ones(nalg,1);
            for r=1:d
                xp = xp + C(r+1,j+1)*Xkj{r};              
            end
        
            % Append collocation equations
            [fj, qj] = f(Xkj{j}, Zkj{j}, Uk);
            g = {g{:}, h*fj - [xp; zp]};
            lbg = [lbg; zeros(nx,1)];
            ubg = [ubg; zeros(nx,1)];

            % Add contribution to the end state
            Xk_end = Xk_end + D(j+1)*Xkj{j};
                        
            % Add contribution to quadrature function
            J = J + B(j+1)*qj*h;
        end

        % New NLP variable for state at end of interval
        % Differential states
        Xk = MX.sym(['X_' num2str(k+1)], ndiff);
        w = {w{:}, Xk};
        lbw = [lbw; xlb];
        ubw = [ubw; xub];
        w0 = [w0; x_init];    
        
        % Add inequality constraint
        g = {g{:}, Uk(6) + Uk(4)*rho*cp*(Xkj{j}(4)-T2)};
        lbg = [lbg; Q_demand((i-1)+k+1)];
        ubg = [ubg; Inf];
            
        % Add equality constraint
        g = {g{:}, Xk_end - Xk};
        lbg = [lbg; zeros(ndiff,1)];
        ubg = [ubg; zeros(ndiff,1)];
    end

    % Create an NLP solver
    opts = struct;
    opts.ipopt.max_iter = maxiter;%5000;
    opts.ipopt.print_level = 5; %0,3
    opts.print_time = 1; %0,1
    opts.ipopt.tol = tol;
    opts.ipopt.acceptable_tol = 100*tol; % optimality convergence tolerance

    prob = struct('f', J, 'x', vertcat(w{:}), 'g', vertcat(g{:}));
    solver = nlpsol('solver', 'ipopt', prob, opts);

    % Solve the NLP
    sol = solver('x0', w0, 'lbx', lbw, 'ubx', ubw,'lbg', lbg, 'ubg', ubg);  
    i_infeasible(i) = solver.stats.success;
    solver_return{i} = solver.stats.return_status; 
    w_opt = full(sol.x);

    if solver.stats.success == 0
        error('Error: Optimal Solution Not Found') 
    end
    
    % Simulator
    x0 = x_init;
    tspan = [0 3600];
    p = [w_opt(nx+1:nx+nu-1); T1; T2; V_hex; V_tank; U_hex; A_hex; rho;... 
        cp; h_s; A_tank; T_s; n];
    options = odeset('RelTol',1e-5,'Stats','off','OutputFcn',@odeplot);
    [t,x] = ode15s(@(t,x) twoPlantModelChen(t,x,p),tspan,x0,options);

    u_opt(:,i) = w_opt(nx+1:nx+nu);
    u_init = w_opt(nx+1:nx+nu);
    x_m = x(end,:);
    x_opt(:,i) = x_m;
    x_init = transpose(x_m);

    x_init
    u_init
    i
end

% Store data
idealNMPC = struct('Demand',Q_demand,'OptimalStates',...
                    x_opt,'OptimalInputs',u_opt,'Measurement',...
                    x_opt,'SolutionStatus',i_infeasible);

%% Plot results
T = T/3600;
tgrid = linspace(0, T, N+1);
clf;

x_opt = [x_0,x_opt];

% Plot state trajectory
figure(1)
plot(tgrid, x_opt(1,:), '-')
hold on
plot(tgrid, x_opt(2,:), '-')
plot(tgrid, x_opt(3,:), '-')
plot(tgrid, x_opt(4,:), '-')
plot(tgrid, x_opt(5,:), '-')
xlabel('t [hr]')
ylabel('T [\circ C]')
legend('T_{L1}','T_{R1}','T_{L2}','T_{R2}','T_{tank}')
hold off
grid on

%Plot optimal controls
figure(2)
subplot(211)
stairs(tgrid, [u_opt(1,:), nan], '-.')
hold on
stairs(tgrid, [u_opt(2,:), nan], '-.')
stairs(tgrid, [u_opt(3,:), nan], '-.')
stairs(tgrid, [u_opt(4,:), nan], '-.')
hold off
xlabel('t [hr]')
ylabel('q_{i} [m^3/s]')
legend('q_{L1}','q_{R2}','q_{L2}','q_{R2}')
grid on

subplot(212)
stairs(tgrid, [u_opt(5,:), nan], '-.')
hold on
stairs(tgrid, [u_opt(6,:), nan], '-.')
hold off
xlabel('t [hr]')
ylabel('Energy supply rate [kW]')
legend('Q_{tank}','Q_M')
grid on

% Plot demand profile
% figure(3)
% stairs(tgrid, Q_demand, '-')
% hold on
% plot(tgrid+0.5, Q_demand, 'k--')
% plot(tgrid, mean(Q_demand)*ones(size(tgrid,2),1),'r--','LineWidth',1.5)
% hold off
% xlabel('t [hr]')
% ylabel('Energy demand rate [kW]')
% ylim([0 12000])
% grid on