% An implementation of direct collocation to open loop 
% dynamic optimisation of a two plant 
% direct supply without storage using CasADi
%% VARIABLE DEMAND!!!

% Zawadi Mdoe, 2019
% =========================================================================
clear;
clc;
close all;

addpath('C:\Users\DELL\Desktop\Matlab\casadi-windows-matlabR2016a-v3.4.5')
import casadi.*

run Parameters_NoStorage.m

Q_demand = [1500*ones(12,1); 3500*ones(12,1)];
Q_supply = 2500*ones(24,1);

% Degree of interpolating polynomial
d = 3;

% Get collocation points
tau_root = [0 collocation_points(d, 'radau')]; %can be 'legendre'

% Coefficients of the collocation equation
C = zeros(d+1,d+1);

% Coefficients of the continuity equation
D = zeros(d+1, 1);

% Coefficients of the quadrature function
B = zeros(d+1, 1);

% Construct polynomial basis
for j=1:d+1
  % Construct Lagrange polynomials to get the polynomial basis 
  % at the collocation point
  coeff = 1;
  for r=1:d+1
    if r ~= j
      coeff = conv(coeff, [1, -tau_root(r)]);
      coeff = coeff / (tau_root(j)-tau_root(r));
    end
  end
  % Evaluate the polynomial at the final time to get the 
  % coefficients of the continuity equation
  D(j) = polyval(coeff, 1.0);

  % Evaluate the time derivative of the polynomial at all collocation 
  % points to get the coefficients of the continuity equation
  pder = polyder(coeff);
  for r=1:d+1
    C(j,r) = polyval(pder, tau_root(r));
  end

  % Evaluate the integral of the polynomial to get the coefficients 
  % of the quadrature function
  pint = polyint(coeff);
  B(j) = polyval(pint, 1.0);
end

% Time horizon
T = 24*60*60;
ndiff = 2;                 %number of differential states
nalg = 2;                 %number of algebraic states
nu = 3;                 %number of controls
nx = nalg + ndiff;           %total number of states

zub = Inf*ones(nalg,1);
zlb = -Inf*ones(nalg,1);

run SSOptNS.m

% Declare model variables
x1 = SX.sym('x1');
x2 = SX.sym('x2');
x3 = SX.sym('x3');
x4 = SX.sym('x4');
x = [x1; x2];
z = [x3; x4];
u1 = SX.sym('u1');      %q_R2
u2 = SX.sym('u2');      %Q_Market
u3 = SX.sym('u3');      %Q_dump
u = [u1; u2; u3];

% Model equations

xdot =  [(1/V_hex)*(q_L1*(T1-x1) - (h_dot)*(0.5*(x3 + x4))^(1/n));

         (1/V_hex)*(u1*(T2-x2) + (h_dot)*(0.5*(x3 + x4))^(1/n));     
         
        T1 - x2 - x3^(1/n);
        
        x1 - T2 - x4^(1/n)];

% Objective term
L = Pm*u2;                  %no tank heating

% Continuous time dynamics
f = Function('f', {x, z, u}, {xdot, L});

% Control discretization
N = 24; % number of control intervals
M = 24; % number of MPC loops
h = T/N;
period = N;


%% Prepare output variables
x_opt = zeros(ndiff,M);
u_opt = zeros(nu,M);
i_infeasible = zeros(M,1);
solver_return = {};

%% Predicted Demand
Q_demand = [Q_demand; Q_demand];
Q_supply = [Q_supply; Q_supply];

%% MPC loop
for i=1:M
    % Start with an empty NLP
    w={};
    w0 = [];
    lbw = [];
    ubw = [];
    J = 0;
    g={};
    lbg = [];
    ubg = [];

    % "Lift" initial conditions
    % Differential states
    Xk = MX.sym('X0', ndiff);
    w = {w{:}, Xk};
    lbw = [lbw; x_init];
    ubw = [ubw; x_init];
    w0 = [w0; x_init];
    
    % Algebraic states
    Zk = MX.sym('Z0', nalg);
    w = {w{:}, Zk};
    lbw = [lbw; z_init];
    ubw = [ubw; z_init];
    w0 = [w0; z_init];
    
    % Formulate the NLP
    for k=0:N-1
        % New NLP variable for the control
        Uk = MX.sym(['U_' num2str(k)], nu); 
        w = {w{:}, Uk};
        lbw = [lbw; ulb];
        ubw = [ubw; uub];
        w0 = [w0; u_init];
  
        % State at collocation points
        Xkj = {};
        Zkj = {};
        for j=1:d
            % Differential states
            Xkj{j} = MX.sym(['X_' num2str(k) '_' num2str(j)], ndiff);
            w = {w{:}, Xkj{j}};
            lbw = [lbw; xlb];
            ubw = [ubw; xub];
            w0 = [w0; x_init];
            
            % Algebraic states
            Zkj{j} = MX.sym(['Z_' num2str(k) '_' num2str(j)], nalg);
            w = {w{:}, Zkj{j}};
            lbw = [lbw; zlb];
            ubw = [ubw; zub];
            w0 = [w0; z_init];           
        end

        % Loop over collocation points
        Xk_end = D(1)*Xk;
       
        for j=1:d
            % Expression for the state derivative at the collocation point
            xp = C(1,j+1)*Xk;
            zp = zeros(nalg,1);
            for r=1:d
                xp = xp + C(r+1,j+1)*Xkj{r};              
            end
        
            % Append collocation equations
            [fj, qj] = f(Xkj{j}, Zkj{j}, Uk);
            g = {g{:}, h*fj - [xp; zp]};
            lbg = [lbg; zeros(nx,1)];
            ubg = [ubg; zeros(nx,1)];

            % Add contribution to the end state
            Xk_end = Xk_end + D(j+1)*Xkj{j};
                        
            % Add contribution to quadrature function
            J = J + B(j+1)*qj*h;
        end

        % New NLP variable for state at end of interval
        % Differential states
        Xk = MX.sym(['X_' num2str(k+1)], ndiff);
        w = {w{:}, Xk};
        lbw = [lbw; xlb];
        ubw = [ubw; xub];
        w0 = [w0; x_init];    
        
        % Add inequality constraint
        g = {g{:}, Uk(2) + Uk(1)*rho*cp*(Xk(2)-T2)};
        lbg = [lbg; Q_demand((i-1)+k+1)];
        ubg = [ubg; Q_demand((i-1)+k+1)];
        
        g = {g{:}, Uk(3) + q_L1*rho*cp*(T1-Xk(1))};
        lbg = [lbg; Q_supply(i)];
        ubg = [ubg; Q_supply(i)];
            
        % Add equality constraint
        g = {g{:}, Xk_end - Xk};
        lbg = [lbg; zeros(ndiff,1)];
        ubg = [ubg; zeros(ndiff,1)];
    end

    % Create an NLP solver
    opts = struct;
    opts.ipopt.max_iter = maxiter;%5000;
    opts.ipopt.print_level = 5; %0,3
    opts.print_time = 1; %0,1
    opts.ipopt.tol = tol;
    opts.ipopt.acceptable_tol = 100*tol; % optimality convergence tolerance

    prob = struct('f', J, 'x', vertcat(w{:}), 'g', vertcat(g{:}));
    solver = nlpsol('solver', 'ipopt', prob, opts);

    % Solve the NLP
    sol = solver('x0', w0, 'lbx', lbw, 'ubx', ubw,'lbg', lbg, 'ubg', ubg);  
    i_infeasible(i) = solver.stats.success;
    solver_return{i} = solver.stats.return_status; 
    w_opt = full(sol.x);

    if solver.stats.success == 0
        error('Error: Optimal Solution Not Found') 
    end
    
    % Simulator
    x0 = x_init;
    tspan = [0 3600];
    p = [q_L1; w_opt(nx+1:nx+nu-2); T1; T2; V_hex; U_hex; A_hex; rho;... 
        cp; n];
    options = odeset('RelTol',1e-5,'Stats','off','OutputFcn',@odeplot);
    [t,x] = ode15s(@(t,x) twoPlantModelDirect(t,x,p),tspan,x0,options);

    u_opt(:,i) = w_opt(nx+1:nx+nu);
    u_init = w_opt(nx+1:nx+nu);
    x_m = x(end,:);
    x_opt(:,i) = x_m(1:ndiff);
    x_init = transpose(x_m(1:ndiff));
    
    x_init
    u_init
    i
end

Cost = Pm*u_opt(2,:)+ Pu*u_opt(3,:);
Cost = [Cost, NaN];

% Store data
nostorageNMPC = struct('Demand',Q_demand,'OptimalStates',...
                    x_opt,'OptimalInputs',u_opt,'Measurement',...
                    x_opt,'Cost',Cost);

%% Plot results
T = T/3600;
tgrid = linspace(0, T, N+1);
clf;

x_opt = [x_0, x_opt];

set(0,'DefaultTextFontName','Times',...
'DefaultTextFontSize',15,...
'DefaultAxesFontName','Times',...
'DefaultAxesFontSize',15,...
'DefaultLineLineWidth',1.5,...
'DefaultLineMarkerSize',7.75,...
'DefaultStairLineWidth',1.5);
set(findall(gcf,'Type','text'),'FontSize',15,'Interpreter','latex');
set(gcf,'color','white');

figure(1)
subplot(311)
plot(tgrid, x_opt(1,:), '-r')
hold on
plot(tgrid, x_opt(2,:), '-b')
% xlabel('time [hr]')
ylabel('Temperature [$\circ$C]','Interpreter','latex')
legend('$T_{1,o}$','$T_{2,o}$','Interpreter','latex')
hold off
grid on

subplot(312)
stairs(tgrid, [u_opt(1,:), nan], '-b')
hold off
% xlabel('time [hr]')
ylabel('$q_{2}$ [$\ell$/s]', 'Interpreter','latex')
ylim([0 10])
grid on

subplot(313)
stairs(tgrid, [u_opt(3,:), nan], '-g')
hold on
stairs(tgrid, [u_opt(2,:), nan], '-r')
hold off
xlabel('time [hr]','Interpreter','latex')
ylabel('Power [kW]','Interpreter','latex')
legend('$Q_{dump}$','$Q_{M}$','Interpreter','latex')
ylim([0 1200])
grid on
