% An implementation of direct collocation to open loop 
% dynamic optimisation of a two plant 
% Energy Storage System using CasADi
%% VARIABLE DEMAND!!!

% Zawadi Mdoe, 2018
% ============================================================================
clear;
clc;

addpath('C:\Users\DELL\Desktop\Matlab\casadi-windows-matlabR2016a-v3.4.5')
import casadi.*

run Parameters_EnergyStorageCase2.m
load CaliforniaDemand.mat
% load Expecteddemand.mat

% Degree of interpolating polynomial
d = 3;

% Get collocation points
tau_root = [0 collocation_points(d, 'radau')]; %can be 'legendre'

% Coefficients of the collocation equation
C = zeros(d+1,d+1);

% Coefficients of the continuity equation
D = zeros(d+1, 1);

% Coefficients of the quadrature function
B = zeros(d+1, 1);

% Construct polynomial basis
for j=1:d+1
  % Construct Lagrange polynomials to get the polynomial basis 
  % at the collocation point
  coeff = 1;
  for r=1:d+1
    if r ~= j
      coeff = conv(coeff, [1, -tau_root(r)]);
      coeff = coeff / (tau_root(j)-tau_root(r));
    end
  end
  % Evaluate the polynomial at the final time to get the 
  % coefficients of the continuity equation
  D(j) = polyval(coeff, 1.0);

  % Evaluate the time derivative of the polynomial at all collocation 
  % points to get the coefficients of the continuity equation
  pder = polyder(coeff);
  for r=1:d+1
    C(j,r) = polyval(pder, tau_root(r));
  end

  % Evaluate the integral of the polynomial to get the coefficients 
  % of the quadrature function
  pint = polyint(coeff);
  B(j) = polyval(pint, 1.0);
end

% Time horizon
T = 24*60*60;
ndiff = 5;                  %number of differential states
nalg = 4;                   %number of algebraic states
nu = 4;                     %number of controls
nx = nalg + ndiff;          %total number of states

zub = Inf*ones(nalg,1);
zlb = -Inf*ones(nalg,1);

% Declare model variables
x1 = SX.sym('x1');
x2 = SX.sym('x2');
x3 = SX.sym('x3');
x4 = SX.sym('x4');
x5 = SX.sym('x5');
x = [x1; x2; x3; x4; x5];
x6 = SX.sym('x6');
x7 = SX.sym('x7');
x8 = SX.sym('x8');
x9 = SX.sym('x9');
z = [x6; x7; x8; x9];
u1 = SX.sym('u1');
u2 = SX.sym('u2');
u3 = SX.sym('u3');
u4 = SX.sym('u4');
u = [u1; u2; u3; u4];

% Model equations

xdot =  [(1/V_hex)*(q_L1*(T1-x1) - h_dot*(0.5*(x6 + x7))^(1/n));

        (1/V_hex)*(u1*(x5-x2) + h_dot*(0.5*(x6 + x7))^(1/n));

        (1/V_hex)*(u2*(x5-x3) - h_dot*(0.5*(x8 + x9))^(1/n));

        (1/V_hex)*(q_R2*(T2-x4) + h_dot*(0.5*(x8 + x9))^(1/n));
        
        (1/V_tank)*(u1*(x2-x5) + u2*(x3-x5) + (u3/(rho*cp)-h_t_dot*(x5-T_s)));
        
        T1-x2-x6^(1/n);
        
        x1-x5-x7^(1/n);
        
        x3-T2-x8^(1/n);
        
        x5-x4-x9^(1/n)];

% Objective term
L = Pm*u4 + Pt*u3 +Px*(x1-80)^2;
% L = Pm*u4 + Pt*u3 + Pu*(u1^2+ u2^2+ u3^2+ u4^2);

% Continuous time dynamics
f = Function('f', {x, z, u}, {xdot, L});

% Control discretization
N = 24; % number of control intervals
M = 24; % number of MPC loops
h = T/N;
period = N;
Q_demand = [CalDemand;CalDemand];

% Variable counting
NXD = N*ndiff*(d+1);    %Total Number of differential state variables 
NXA = N*nalg*d;         %Total Number of algebraic state variables
NU = N*nu;              %Total Number of input variables
NXF = ndiff;            %Total Number of end point variables (differential only)
NV = NXD + NXA + NU + NXF;  %Total number of NLP variables

%% Prepare output variables
x_opt = zeros(ndiff,M);
u_opt = zeros(nu,M);
i_infeasible = zeros(M,1);
solver_return = {};


% Start with an empty NLP
w={};
w0 = [];
lbw = [];
ubw = [];
J = 0;
g={};
lbg = [];
ubg = [];

% "Lift" initial conditions
% Differential states
Xk = MX.sym('X0', ndiff);
w = {w{:}, Xk};
lbw = [lbw; x_init];
ubw = [ubw; x_init];
w0 = [w0; x_init];
      
% Formulate the NLP
for k=0:N-1
    % New NLP variable for the control
    Uk = MX.sym(['U_' num2str(k)],nu); 
    w = {w{:}, Uk};
    lbw = [lbw; ulb];
    ubw = [ubw; uub];
    w0 = [w0; u_init];
  
    % State at collocation points
    Xkj = {};
    Zkj = {};
        
    for j=1:d
        % Differential states
        Xkj{j} = MX.sym(['X_' num2str(k) '_' num2str(j)],ndiff);
        w = {w{:}, Xkj{j}};
        lbw = [lbw; xlb];
        ubw = [ubw; xub];
        w0 = [w0; x_init];
            
        % Algebraic states
        Zkj{j} = MX.sym(['Z_' num2str(k) '_' num2str(j)],nalg);
        w = {w{:}, Zkj{j}};
        lbw = [lbw; zlb];
        ubw = [ubw; zub];
        w0 = [w0; z_init];           
    end

    % Loop over collocation points
    Xk_end = D(1)*Xk;
        
    for j=1:d
        % Expression for the state derivative at the collocation point
        xp = C(1,j+1)*Xk;
        for r=1:d
            xp = xp + C(r+1,j+1)*Xkj{r};              
        end
            
        % Append collocation equations
        [fj, qj] = f(Xkj{j},Zkj{j},Uk);          
        g = {g{:}, h*fj - [xp; zeros(nalg,1)]};
        lbg = [lbg; zeros(nx,1)];
        ubg = [ubg; zeros(nx,1)];
            
        % Add contribution to the end state
        Xk_end = Xk_end + D(j+1)*Xkj{j};
                        
        % Add contribution to quadrature function
        J = J + B(j+1)*qj*h;
    end
            
    % New NLP variable for state at end of interval
    % Differential states
    Xk = MX.sym(['X_' num2str(k+1)], ndiff);
    w = {w{:}, Xk};
    lbw = [lbw; xlb];
    ubw = [ubw; xub];
    w0 = [w0; x_init];
        
    % Add inequality constraints
    g = {g{:}, Uk(4) + Uk(2)*rho*cp*(Xk(4)-T2)};
    lbg = [lbg; Q_demand(k+1)];
    ubg = [ubg; Q_demand(k+1)];
                      
    % Add equality constraint
    g = {g{:}, Xk_end-Xk};
    lbg = [lbg; zeros(ndiff,1)];
    ubg = [ubg; zeros(ndiff,1)];
    
end

% Create an NLP solver
opts = struct;
opts.ipopt.max_iter = maxiter;
opts.ipopt.print_level = 5; %0,3
opts.print_time = 1; %0,1
opts.ipopt.tol = tol;
opts.ipopt.acceptable_tol =100*tol; % optimality convergence tolerance

prob = struct('f', J, 'x', vertcat(w{:}), 'g', vertcat(g{:}));
solver = nlpsol('solver', 'ipopt', prob, opts);

% Solve the NLP
sol = solver('x0', w0, 'lbx', lbw, 'ubx', ubw,'lbg', lbg, 'ubg', ubg);
w_opt = full(sol.x);

% Store data
OLhigherorder = struct('Demand',Q_demand,'OptimalSolution',...
                    w_opt);

%% Plot results
T = T/3600;
tgrid = linspace(0, T, N+1);
clf;

%% Plot the solution
x1_opt = w_opt(1:(ndiff+nu)+nx*d:end);
x2_opt = w_opt(2:(ndiff+nu)+nx*d:end);
x3_opt = w_opt(3:(ndiff+nu)+nx*d:end);
x4_opt = w_opt(4:(ndiff+nu)+nx*d:end);
x5_opt = w_opt(5:(ndiff+nu)+nx*d:end);
u1_opt = w_opt(6:(ndiff+nu)+nx*d:end);
u2_opt = w_opt(7:(ndiff+nu)+nx*d:end);
u3_opt = w_opt(8:(ndiff+nu)+nx*d:end);
u4_opt = w_opt(9:(ndiff+nu)+nx*d:end);

% Plot state trajectory
figure(1)
plot(tgrid, x1_opt, '-')
hold on
plot(tgrid, x2_opt, '-')
plot(tgrid, x3_opt, '-')
plot(tgrid, x4_opt, '-')
plot(tgrid, x5_opt, '-')
xlabel('t [hr]')
ylabel('T [\circ C]')
legend('T_{L1}','T_{R1}','T_{L2}','T_{R2}','T_{tank}')
hold off
grid on

%Plot optimal controls
figure(2)
subplot(211)
stairs(tgrid, q_L1*ones(N+1,1), 'r-')
hold on
stairs(tgrid, [u1_opt; nan], 'g-')
stairs(tgrid, [u2_opt; nan], 'b-')
stairs(tgrid, q_R2*ones(N+1,1), 'k-')
hold off
xlabel('t [hr]')
ylabel('q_{i} [m^3/s]')
legend('q_{L1}','q_{R1}','q_{L2}','q_{R2}')
grid on

subplot(212)
stairs(tgrid, [u3_opt; nan], '-')
hold on
stairs(tgrid, [u4_opt; nan], '-')
stairs(tgrid, Q_demand(1:size(tgrid,2)), 'k*')
hold off
xlabel('t [hr]')
ylabel('Energy supply rate [kW]')
legend('Q_{tank}'   ,'Q_M')
grid on

% % Plot demand profile
% figure(3)
% stairs(tgrid, Q_demand(1:size(tgrid,2)), '-')
% hold on
% plot(tgrid+0.5, Q_demand(1:size(tgrid,2)), 'k--')
% plot(tgrid, mean(Q_demand)*ones(size(tgrid,2),1),'r--','LineWidth',1.5)
% hold off
% xlabel('t [hr]')
% ylabel('Energy demand rate [kW]')
% ylim([0 12000])
% grid on